# Navigation is a non-database-model.  It represents data and boundaries for
# the user navigating (panning and zooming) around the tree view.  This class
# allows you to specify a zoom, and will determine the approrpriate viewBox
# size for the SVG picture.  As well, for a given zoom, it will provide
# boundaries for panning around the tree.
#
# This class also renders SVG "zoom in" and "zoom out" buttons
class Navigation
  attr :zoom
  attr :viewwidth
  attr :viewheight
  attr :totalwidth
  attr :totalheight

  # The most that the user can zoom in by.  The least is always 1 (you can't
  # zoom out beyond seeing everything at 1x).
  MAX_ZOOM = 10

  # Create a new Navigation object. It is initialized to a 1x zoom.
  def initialize()
    @zoom = 1
    @totalwidth = 1
    @totalheight = 1
    @viewx = 0
    @viewy = 0
    @viewwidth = 1
    @viewheight = 1
  end

  # Set the size of the SVG picture being viewed
  # @param w The width of the picture
  # @param h The height of the picture
  def set_size(w, h)
    @totalwidth = w
    @totalheight = h

    # adjust the view size
    @viewwidth = @totalwidth / @zoom
    @viewheight = @totalheight / @zoom

    # check bounds for the position
    position_verify
  end

  # Change the zoom
  # @param z The magnification requested (2 means 2x, etc)
  def zoom=(z)
    return if z < 1 or z > MAX_ZOOM # upper and lower limits

    vw = @totalwidth / z
    vh = @totalheight / z

    # center the view for the new zoom (this will be verified later so don't
    # worry if it's valid right now)
    @viewx += (@viewwidth - vw) / 2
    @viewy += (@viewheight - vh) / 2

    # apply the new zoom
    @viewwidth = vw
    @viewheight = vh
    @zoom = z

    # check bounds for the position
    position_verify
  end

  # Change the left edge of the current view
  # @param x The new left edge
  def viewx=(x)
    @viewx = x
    # check bounds for the position
    position_verify
  end

  # Change the top edge of the current view
  # @param y The new top edge
  def viewy=(y)
    @viewy = y
    # check bounds for the position
    position_verify
  end

  # The left edge of the current view
  def viewx
    position_verify
    @viewx
  end

  # The top edge of the current view
  def viewy
    position_verify
    @viewy
  end

  # The left boundary for the left edge at the current zoom
  def min_viewx
    0
  end

  # The top boundary for the top edge at the current zoom
  def min_viewy
    0
  end

  # The right boundary for the left edge at the current zoom
  def max_viewx
    @totalwidth - @viewwidth
  end

  # The bottom boundary for the top edge at the current zoom
  def max_viewy
    @totalheight - @viewheight
  end

  # Render "zoom in" and "zoom out" buttons to XML
  # @param root A libXML <svg> node which the zoom buttons will be inserted
  #             under as children.
  def render_svg(root)
    puts @totalwidth, @totalheight

    zoomin = XML::Node.new('svg')
    zoomin['id'] = 'zoom-in'
    zoomin['x'] = '0'
    zoomin['y'] = '0'
    zoomin['width'] = '18'
    zoomin['height'] = '18'
    zoomin.child = p = XML::Node.new('path')
    p['d'] = "M0,0 L17,0 L17,17 L0,17 Z M4,8.5 L13,8.5 M8.5,4 L8.5,13"
    p['stroke-width'] = '2'
    p['stroke'] = '#000000'
    p['fill'] = '#ffffff'

    zoomout = XML::Node.new('svg')
    zoomout['id'] = 'zoom-out'
    zoomout['x'] = '20'
    zoomout['y'] = '0'
    zoomout['width'] = '18'
    zoomout['height'] = '18'
    zoomout.child = p = XML::Node.new('path')
    p['d'] = "M0,0 L17,0 L17,17 L0,17 Z M4,8.5 L13,8.5"
    p['stroke-width'] = '2'
    p['stroke'] = '#000000'
    p['fill'] = '#ffffff'

    root.child = zoomin
    root.child = zoomout
  end

protected
  def position_verify()
    @viewx = (@viewx < min_viewx ? min_viewx :
              (@viewx > max_viewx ? max_viewx : @viewx))
    @viewy = (@viewy < min_viewy ? min_viewy :
              (@viewy > max_viewy ? max_viewy : @viewy))
  end
end
