require 'naerrors'
gem 'libxml-ruby'
require 'xml/libxml'

# A Tree class represents a tree in the database.  Only one revision of a tree
# can be loaded into a Tree object at a time.
# A Tree also can render itself into an SVG picture.
class Tree < ActiveRecord::Base
  attr :root
  attr :revision

  # Scale everything by this
  SCALE=1
  # The minimum zoom for a node
  MIN_ZOOM=0.5
  # The maximim zoom for a node
  MAX_ZOOM=1.5
  # The base width of a node
  BASE_NODE_SIZE = 100 * SCALE
  # The amount of padding between things vertically
  VERT_PADDING = 24 * SCALE
  # The amount of padding between things horizontally
  HORZ_PADDING = 24 * SCALE

  # Load a revision of this tree into the object.  After calling this, the
  # root attribute will point to the root node of the loaded tree.
  # @param revision The ID of the revision to be loaded
  def load_revision(revision)
    node_cache_init(revision)
    begin
      @root = load_nodes(1)
    rescue NetAdhict::PositionNotFound
      @root = nil
    end
  end

  # Find a node in the tree by its unique id
  # @param id The unique id of a node
  # @return Returns the InternalNode or TerminalNode which matches the
  #         the given unique id
  def find_node(id)
    def find(node, id)
      return nil if !node
      return node if (node.node_id == id)
      l = find(node.left, id)
      return l if l
      return find(node.right, id)
    end
    find(@root, id)
  end

  # Returns a list of the unique ids for every node (terminal and internal) in
  # the tree
  def all_node_ids
    def get_id(n)
      return [] if !n
      [n.node_id].concat(get_id(n.left)).concat(get_id(n.right))
    end
    get_id(@root)
  end

  # Returns a list of the unique ids for every terminal node in the tree
  def all_terminal_node_ids
    def get_id(n)
      return [n.node_id] if n.terminal
      get_id(n.left).concat(get_id(n.right))
    end
    get_id(@root)
  end

  # Builds the SVG representation of the tree (not the final XML
  # representation)
  # @param A hash of TrafficSummaries for each unique node id in the tree.
  def build_svg(traffics)
    return if !@root

    # find all the terminal nodes and the max traffic summary total
    def max_traffic(n, traffics)
      return 0 if !n
      l = max_traffic(n.left, traffics)
      r = max_traffic(n.right, traffics)
      if l > r then max = l else max = r end
      if n.terminal then my = traffics[n.node_id].total else my = 0 end
      if my > max then max = my end
      max
    end
    max = max_traffic(@root, traffics)
    max = 1 if max == 0 # don't allow a zero max (division by zero!)
    build_nodes(@root, traffics, max)
  end

  # Layout the SVG representation of the tree into a tree-like thing
  def layout_svg()
    return [0, 0] if !@root

    rows = []   # y position of each row
    rowsh = []  # height of each row
    leftedges = []

    rows[0] = 0
    layout_find_rows([@root], rows, rowsh, 0)
    @totalheight = rows[-1] + rowsh[-1]

    @totalwidth = layout_nodes(@root)

    layout_lines(@root, nil)

    [@totalwidth, @totalheight]
  end

  # Render the SVG picture of the tree into its XML representation
  # @return Returns 2 values. Firstly, returns the libXML object for the
  #         SVG root of the picture.  Secondly, returns the libXML object
  #         for the SVG root of the actual tree itself.  The tree is a child
  #         element of the overall picture.
  def render_svg()
    doc = XML::Document.new
    doc.root = XML::Node.new('svg')
    doc.root['id'] = 'svg-picture-root' # for the AJAX to find it
    doc.root['xmlns'] = 'http://www.w3.org/2000/svg'
    doc.root['xmlns:xlink'] = 'http://www.w3.org/2000/xlink/namespace/'
    doc.root['version'] = '1.1'
    doc.root['width'] = '100%'
    doc.root['height'] = '100%'

    tree_root = XML::Node.new('svg')
    tree_root['id'] = 'svg-tree-root' # for the AJAX to find it
    tree_root['x'] = '0'
    tree_root['y'] = '0'
    tree_root['width'] = '100%'
    tree_root['height'] = '100%'
    doc.root.child = tree_root

    render_nodes(@root, tree_root) if @root

    #p = create_pie(root.id)
    #s = p.add_slice(SVG::Pie::RING_TCPIP, 100, Colors.slice(:HTTP))

    #doc.root.child = n.to_xml()

    [doc.root, tree_root]
  end

protected
  def node_cache_init(revision_id)
    @ncache = {}
    if revision_id
      cond = "revision_id=#{revision_id.to_s}"

      InternalNode.find(:all, :conditions => cond).each {|n|
        @ncache[n.position] = n
      }
      TerminalNode.find(:all, :conditions => cond).each {|n|
        @ncache[n.position] = n
      }
    end
  end

  def node_cache(position)
    @ncache[position]
  end

  # find all nodes under this position recursively
  def load_nodes(position)

    if (!(node = node_cache(position)))
      raise NetAdhict::PositionNotFound,
        "No position #{position.to_s} found in tree"
    end

    if (!node.terminal)
      l = load_nodes(position * 2)
      r = load_nodes(position * 2 + 1)
      node.left = l
      node.right = r
      l.parent = node
      r.parent = node

      node.zoom = 1 # internal nodes are always zoom 1?
      #node.collapsed = false
    else
      node.zoom = 1 # this can change for terminal nodes
    end

    node
  end

  def create_ring_color(label)
    c = Classification.find_label(label)
    SVG::Color.new(c.color)
  end

  def create_pie(n)
    p = SVG::Pie.new(n.node_id, n.zoom * BASE_NODE_SIZE)
    p.rings[SVG::Pie::RING_IP].color = create_ring_color('ether_ipv4')
    p.rings[SVG::Pie::RING_TCPIP].color = create_ring_color('ether_ipv4_tcp')
    p.rings[SVG::Pie::RING_UDPIP].color = create_ring_color('ether_ipv4_udp')
    p
  end

  def create_line(id)
    l = SVG::Line.new(id)
  end

  def create_slice(pie, percent, label, color, ring)
    s = pie.add_slice(ring, percent, SVG::Color.new(color))
    s.tooltip = label
    s
  end

  def create_node(n)
    p = SVG::Node.new(n.node_id,
                      n.zoom * BASE_NODE_SIZE,
                      n.zoom * BASE_NODE_SIZE * 3 / 4)
  end

  def layout_find_rows(nodesatlevel, rows, rowsh, level)
    return if nodesatlevel.empty?

    if (level > 0)
      rows[level] = rows[level - 1] + rowsh[level - 1];
    end

    thisrowheight = 0
    nodesatnextlevel = []
    nodesatlevel.each {|n|
      # position them vertically
      n.svg.y = rows[level]

      thisrowheight = n.svg.height if n.svg.height > thisrowheight
      nodesatnextlevel << n.left << n.right if (!n.terminal)
    }
    thisrowheight += VERT_PADDING
    rowsh[level] = thisrowheight

    # go to the next row
    layout_find_rows(nodesatnextlevel, rows, rowsh, level+1)
  end

  # returns the highest level that was moved
  def shift_node(n, amt, level, lowest)
    mid = n.svg.x + amt + n.svg.width / 2
    parentmid = n.parent.svg.x + n.parent.svg.width / 2

    # move the parent if we're moving left and we're the right child, or
    # if we're moving right and we're the left child
    moveparent_amt = mid - parentmid
    if !((moveparent_amt > 0 && n.leftchild?) ||
         (moveparent_amt < 0 && n.rightchild?))
      moveparent_amt = 0
    end

    amt -= moveparent_amt
    n.svg.x += amt
    find_sides(n)

    # move the whole subtree by the same amount
    row = [n]
    (l = level).upto(lowest) do |i|
      nextrow = []
      row.each do |m| nextrow << m.left << m.right if !m.terminal end
      row = nextrow

      row.each do |m| m.svg.x += amt; find_sides(m) end
    end

    if moveparent_amt != 0
      return shift_node(n.parent, moveparent_amt, level - 1, lowest)
    elsif amt != 0
      return level
    else
      return level + 1
    end
  end

  def layout_children(root)
    level_nodes = [[root]]

    # build up our array of all the nodes, with one sub-array per row
    level = 0
    begin
      nextlevel = []
      level_nodes[level].each do |n|
        nextlevel << n.left << n.right if !n.terminal
      end
      level += 1
      level_nodes[level] = nextlevel
    end while !level_nodes[level].empty?

    level = 0
    lowest_level = level
    begin
      # the lowest level down the tree that's been moved this iteration
      level_moved = level + 1
      
      # look for nodes in eachothers' space
      last = nil
      level_nodes[level].each do |n|
        if (last)
          shift_amt = (last.svg.x + last.svg.width - 1) + HORZ_PADDING -
            n.svg.x
          if shift_amt >= 1 # less than 1 pixel, don't worry about
            # collision! do we shift the last node left, or the current one
            # right, or both?
            level_midpoint = (2 ** level + 2 ** (level - 1))
            shift_left = last.position < level_midpoint
            shift_right = !(n.position < level_midpoint)
            # sanity check..
            raise "what the heck?" if !shift_left && !shift_right

            # if we're shifting both ways, then shift each way by 1/2
            shift_amt /= 2 if shift_left && shift_right

            # shift the nodes over and their parents etc
            if shift_left
              m = shift_node(last, -shift_amt, level, lowest_level)
              level_moved = m if m < level_moved
            end
            if shift_right
              m = shift_node(n, shift_amt, level, lowest_level)
              level_moved = m if m < level_moved
            end
          end
        end

        last = n
      end

      # place their children their default spots if we're done placing up
      # til here
      if level_moved > lowest_level
        level_nodes[level_moved].each do |n|
          x = n.parent.svg.x + n.parent.svg.width / 2;
          if n.leftchild?
            x -= (n.svg.width + HORZ_PADDING) / 2
          else
            x += (n.svg.width + HORZ_PADDING) / 2
          end
          x -= n.svg.width / 2
          n.svg.x = x
          find_sides(n)
        end
      end

      # go back to the highest level that was moved, or move on if nothing was
      # moved at all for this level or above
      level = level_moved
      lowest_level = level if level > lowest_level
    end while !level_nodes[level].empty?
  end

  def find_sides(n)
    left = n.svg.x
    right = n.svg.x + n.svg.width - 1
    @leftside = left if left < @leftside
    @rightside = right if right > @rightside
  end

  def layout_nodes(root)
    root.svg.x = -root.svg.width / 2
    @leftside = @rightside = 0
    find_sides(root)
    layout_children(root)

    width = (@rightside - @leftside + 1).round
    offset = -@leftside

    # adjust them all away from the 0 coord
    row = [root]
    while !row.empty?
      nextrow = []
      row.each do |n|
        n.svg.x += offset
        nextrow << n.left << n.right if !n.terminal
      end
      row = nextrow
    end

    return width
  end

  def layout_lines(n, parent)
    # line end pos
    if parent and parent.left == n
      parent.svgleft.x2 = (n.svg.x + n.svg.width / 2)
      parent.svgleft.y2 = n.svg.y
    end
    if parent and parent.right == n
      parent.svgright.x2 = (n.svg.x + n.svg.width / 2)
      parent.svgright.y2 = n.svg.y
    end

    if (!n.terminal)
      # line start pos
      n.svgleft.x1 = n.svgright.x1 = (n.svg.x + n.svg.width / 2)
      n.svgleft.y1 = n.svgright.y1 = (n.svg.y + n.svg.height)

      layout_lines(n.left, n)
      layout_lines(n.right, n)
    end
  end
  
  def build_nodes(n, traffics, max)
    return if (!n)

    if (n.terminal)
      n.svg = build_terminal_node(n, traffics, max)
    else
      n.svg = build_internal_node(n)
      build_nodes(n.left, traffics, max)
      build_nodes(n.right, traffics, max)
      n.svgleft = create_line("#{n.id}-#{n.left.id}")
      n.svgright = create_line("#{n.id}-#{n.right.id}")
    end
  end

  def render_nodes(n, docroot)
    docroot.child = c = n.svg.to_xml
    docroot.child = n.svgleft.to_xml if n.svgleft
    docroot.child = n.svgright.to_xml if n.svgright

    c['onclick'] = "node_click(#{c['id']})"

    if (!n.terminal)
      render_nodes(n.left, docroot)
      render_nodes(n.right, docroot)
    end
  end

  def build_terminal_node(n, traffics, max)
    summary = traffics[n.node_id]

    n.zoom = summary.total * (MAX_ZOOM-MIN_ZOOM) / max + MIN_ZOOM
    p = create_pie(n)

    # do we have user defined labels?
    user_labels = Label.find_for_node(self.id, n)

    summary.names.each do |name|
      ul = user_labels[name]
      classification = Classification.find_label(name)

      ring = case classification.ring
             when 0
               SVG::Pie::RING_NONE
             when 1
               SVG::Pie::RING_IP
             when 2
               SVG::Pie::RING_TCPIP
             when 3
               SVG::Pie::RING_UDPIP
             else
               throw "Unhandled ring in the classifications table"
             end
      color = ul ? ul.color : classification.color
      label = ul ? ul.name : classification.label
      create_slice(p, summary.percent(name), label, color, ring)
    end

    p.text = SVG::Text.new("\n" + n.node_id.to_s)

    p
  end

  def build_internal_node(n)
    p = create_node(n)
    pngs = NodePngram.find_for_node(n)

    str = "#{n.node_id}"
    pngs.each {|png|
      str += "\n(#{png.p}, 0x"
      png.n.each_byte {|c| str += "%02x" % c}
      str += ')'
    }
    p.text = SVG::Text.new(str)
    p
  end
end
